# Tai Sakuma <tai.sakuma@gmail.com>
import sys

from .base import Presentation

##__________________________________________________________________||
class ProgressBar(Presentation):
    def __init__(self):
        super(ProgressBar, self).__init__()
        self.lines = [ ]
        self.interval = 0.1 # [second]

    def __repr__(self):
        return '{}()'.format(
            self.__class__.__name__
        )

    def _present(self):
        self._delete_previous_lines()
        self._create_lines()
        self._print_lines()

    def _delete_previous_lines(self):
        if len(self.lines) >= 1:
            sys.stdout.write('\b'*len(self.lines[-1]))
        if len(self.lines) >= 2:
            sys.stdout.write('\033M'*(len(self.lines) - 1))
        self.lines = [ ]
        self.last = [ ]

    def _create_lines(self):
        for taskid in self._active_taskids + self._new_taskids:
            report = self._report_dict[taskid]
            line = self._create_line(report)
            self.lines.append(line)
        for taskid in self._finishing_taskids:
            report = self._report_dict[taskid]
            line = self._create_line(report)
            self.last.append(line)

    def _create_line(self, report):
        name_field_length = 32
        percent = float(report['done'])/report['total'] if report['total'] > 0 else 1
        bar = (':' * int(percent * 40)).ljust(40, " ")
        percent = round(percent * 100, 2)
        name = report['name'][0:name_field_length]
        format = ' {percent:6.2f}% {bar:s} | {done:8d} / {total:8d} |:  {name:<{name_field_length}s} '
        ret = format.format(
            percent=percent, bar=bar,
            done=report['done'], total=report['total'],
            name=name, name_field_length=name_field_length)
        return ret

    def _print_lines(self):
        if len(self.last) > 0: sys.stdout.write("\n".join(self.last) + "\n")
        sys.stdout.write("\n".join(self.lines))
        sys.stdout.flush()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import os, time
import threading

from .detach import detach_pickup

##__________________________________________________________________||
class ProgressReportPickup(threading.Thread):
    """A pickup of progress reports.

    This class picks up progress reports and presents them.

    """
    def __init__(self, queue, presentation):
        threading.Thread.__init__(self)
        self.queue = queue
        self.presentation = presentation
        self.last_wait_time = 1.0 # [second]
        self.taskids = set()

    def run(self):
        self._run_until_the_end_order_arrives()
        self._run_until_reports_stop_coming()

    def _run_until_the_end_order_arrives(self):
        end_order_arrived = False
        while not end_order_arrived:
            while not self.queue.empty():
                report = self.queue.get()
                if report is None: # the end order
                    end_order_arrived = True
                    continue
                self._process_report(report)

    def _run_until_reports_stop_coming(self):
        self._read_time()
        while self.presentation.active():
            if self._time() - self.last_time > self.last_wait_time:
                break
            while not self.queue.empty():
                report = self.queue.get()
                if report is None:
                    continue
                self._read_time()
                self._process_report(report)

    def _process_report(self, report):
        self._detach_self_if_necessary(report)
        self.presentation.present(report)

    def _detach_self_if_necessary(self, report):
        if report['taskid'] in self.taskids:
            return
        self.taskids.add(report['taskid'])
        if os.getpid() == report['pid'] and report['in_main_thread']:
            return
        detach_pickup()

    def _time(self):
        return time.time()

    def _read_time(self):
        self.last_time = self._time()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import time
import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

from atpbar.presentation.base import Presentation

##__________________________________________________________________||
class MockProgressBar(Presentation):
    def _present(self):
        pass

##__________________________________________________________________||
@pytest.fixture()
def mock_time(monkeypatch):
    ret = mock.Mock()
    monkeypatch.setattr(time, 'time', ret)
    ret.return_value = 1533374055.904203
    return ret

@pytest.fixture()
def obj(mock_time):
    return MockProgressBar()

##__________________________________________________________________||
def test_repr(obj):
    repr(obj)

##__________________________________________________________________||
def test_present(obj):
    obj.present(dict(taskid=1, last=False))
    assert obj.active()
    obj.present(dict(taskid=1, last=False))
    assert obj.active()
    obj.present(dict(taskid=2, last=False))
    assert obj.active()
    obj.present(dict(taskid=2, last=False))
    assert obj.active()
    obj.present(dict(taskid=1, last=True))
    assert obj.active()
    obj.present(dict(taskid=2, last=True))
    assert not obj.active()

##__________________________________________________________________||
params = [

    ##
    pytest.param(
        dict(taskid=1, last=False),
        [ ], [ ], [ ], [ ],
        [1], [ ], [ ], [ ],
        True
    ),
    pytest.param(
        dict(taskid=1, last=False),
        [1], [ ], [ ], [ ],
        [1], [ ], [ ], [ ],
        True
    ),
    pytest.param(
        dict(taskid=1, last=False),
        [ ], [1], [ ], [ ],
        [ ], [1], [ ], [ ],
        True
    ),
    pytest.param(
        dict(taskid=1, last=False),
        [ ], [ ], [1], [ ],
        [ ], [ ], [1], [ ],
        True
    ),
    pytest.param(
        dict(taskid=1, last=False),
        [ ], [ ], [ ], [1],
        [ ], [ ], [ ], [1],
        False
    ),

    ##
    pytest.param(
        dict(taskid=1, last=True),
        [ ], [ ], [ ], [ ],
        [ ], [ ], [1], [ ],
        True
    ),
    pytest.param(
        dict(taskid=1, last=True),
        [1], [ ], [ ], [ ],
        [ ], [ ], [1], [ ],
        True
    ),
    pytest.param(
        dict(taskid=1, last=True),
        [ ], [1], [ ], [ ],
        [ ], [ ], [1], [ ],
        True
    ),
    pytest.param(
        dict(taskid=1, last=True),
        [ ], [ ], [1], [ ],
        [ ], [ ], [1], [ ],
        True
    ),
    pytest.param(
        dict(taskid=1, last=True),
        [ ], [ ], [ ], [1],
        [ ], [ ], [ ], [1],
        False
    ),
]
param_names = (
    'report, '
    'initial_new_taskids, initial_active_taskids, '
    'initial_finishing_taskids, initial_complete_taskids, '
    'expected_new_taskids, expected_active_taskids, '
    'expected_finishing_taskids, expected_complete_taskids, '
    'expected_return'
)

@pytest.mark.parametrize(param_names, params)
def test_register_report(
        obj, report,
        initial_new_taskids, initial_active_taskids,
        initial_finishing_taskids, initial_complete_taskids,
        expected_new_taskids, expected_active_taskids,
        expected_finishing_taskids, expected_complete_taskids,
        expected_return):

    obj._new_taskids[:] = initial_new_taskids
    obj._active_taskids[:] = initial_active_taskids
    obj._finishing_taskids[:] = initial_finishing_taskids
    obj._complete_taskids[:] = initial_complete_taskids

    assert expected_return == obj._register_report(report)

    assert expected_new_taskids == obj._new_taskids
    assert expected_active_taskids == obj._active_taskids
    assert expected_finishing_taskids == obj._finishing_taskids
    assert expected_complete_taskids == obj._complete_taskids

##__________________________________________________________________||
params = [
    pytest.param(
        [ ], [ ], [ ], [ ],
        [ ], [ ], [ ], [ ],
    ),
    pytest.param(
        [ ], [10, 11, 12], [ ], [30, 31, 32],
        [ ], [10, 11, 12], [ ], [30, 31, 32],
    ),
    pytest.param(
        [1, 2], [10, 11, 12], [20, 21], [30, 31, 32],
        [ ], [10, 11, 12, 1, 2], [ ], [30, 31, 32, 20, 21],
    ),
]
param_names = (
    'initial_new_taskids, initial_active_taskids, '
    'initial_finishing_taskids, initial_complete_taskids, '
    'expected_new_taskids, expected_active_taskids, '
    'expected_finishing_taskids, expected_complete_taskids, '
)

@pytest.mark.parametrize(param_names, params)
def test_update_registry(
        obj,
        initial_new_taskids, initial_active_taskids,
        initial_finishing_taskids, initial_complete_taskids,
        expected_new_taskids, expected_active_taskids,
        expected_finishing_taskids, expected_complete_taskids):

    obj._new_taskids[:] = initial_new_taskids
    obj._active_taskids[:] = initial_active_taskids
    obj._finishing_taskids[:] = initial_finishing_taskids
    obj._complete_taskids[:] = initial_complete_taskids

    obj._update_registry()

    assert expected_new_taskids == obj._new_taskids
    assert expected_active_taskids == obj._active_taskids
    assert expected_finishing_taskids == obj._finishing_taskids
    assert expected_complete_taskids == obj._complete_taskids

##__________________________________________________________________||
params = [
    pytest.param([ ], [ ], 4.0, 2.0, 1.0, True),
    pytest.param([1], [ ], 4.0, 2.0, 1.0, True),
    pytest.param([ ], [1], 4.0, 2.0, 1.0, True),
    pytest.param([ ], [ ], 4.0, 2.0, 3.0, False),
]
param_names = (
    'new_taskids, finishing_taskids, '
    'current_time, last_time, interval, expected'
)

@pytest.mark.parametrize(param_names, params)
def test_need_to_present(
        obj, mock_time, new_taskids, finishing_taskids,
        current_time, last_time, interval, expected):

    obj._new_taskids[:] = new_taskids
    obj._finishing_taskids[:] = finishing_taskids

    mock_time.return_value = current_time
    obj.last_time = last_time
    obj.interval = interval

    assert expected == obj._need_to_present()

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>

##__________________________________________________________________||
to_detach_pickup = False

##__________________________________________________________________||
def detach_pickup():
    """detach the pickup from any instances of `atpbar`

    If a pickup is started by an `atpbar` in the main thread of the
    main process, the pickup will be owned by the `atpbar`. The
    `atpbar` will end the pickup when the loop ends. This function
    detaches the pickup from the `atpbar` and prevents the atpbar from
    ending the pickup.

    This function is typically called by the pickup when the pickup
    receives a report from a sub-thread or a sub-process.

    """

    global to_detach_pickup
    to_detach_pickup = True
     # Locking here would causes a deadlock. `flush()` locks while
     # ending the pickup. Before receiving the end order, the pickup
     # might still receives a report with a new task ID from a
     # sub-thread or sub-process, it will call this function and will
     # cause a deadlock.


##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys

import pytest

try:
    import unittest.mock as mock
except ImportError:
    import mock

has_jupyter_notebook = False
try:
    import ipywidgets as widgets
    from IPython.display import display
    has_jupyter_notebook = True
except ImportError:
    pass

from atpbar.presentation.create import create_presentation

##__________________________________________________________________||
@pytest.fixture(
    params=[True, False]
)
def isatty(request, monkeypatch):
    ret = request.param
    org_stdout = sys.stdout
    f = mock.Mock(**{
        'stdout.isatty.return_value': ret,
        'stdout.write.side_effect': lambda x : org_stdout.write(x)
    })
    module = sys.modules['atpbar.presentation.create']
    monkeypatch.setattr(module, 'sys', f)
    return ret

##__________________________________________________________________||
if has_jupyter_notebook:
    is_jupyter_notebook_parames = [True, False]
else:
    is_jupyter_notebook_parames = [False]

@pytest.fixture(params=is_jupyter_notebook_parames)
def is_jupyter_notebook(request, monkeypatch):
    ret = request.param
    f = mock.Mock()
    f.return_value = ret
    module = sys.modules['atpbar.presentation.create']
    monkeypatch.setattr(module, 'is_jupyter_notebook', f)
    return ret

##__________________________________________________________________||
def test_create_presentation(isatty, is_jupyter_notebook):
    actual = create_presentation()

    if isatty:
        assert 'ProgressBar' == actual.__class__.__name__
    elif is_jupyter_notebook:
        assert 'ProgressBarJupyter' == actual.__class__.__name__
    else:
        assert 'ProgressPrint' == actual.__class__.__name__

##__________________________________________________________________||



# Tai Sakuma <tai.sakuma@gmail.com>
import sys
import time, random
import itertools
import threading
import multiprocessing

import pytest

from atpbar import atpbar
from atpbar import register_reporter, find_reporter, flush, disable

##__________________________________________________________________||
@pytest.mark.parametrize('niterations', [10, 1, 0])
def test_one_loop(mock_create_presentation, wrap_end_pickup, niterations):

    #
    disable()

    #
    for i in atpbar(range(niterations)):
        pass

    print()
    print(mock_create_presentation)

    #
    assert 0 == wrap_end_pickup.call_count

    #
    presentations = mock_create_presentation.presentations
    assert 0 == len(presentations)


##__________________________________________________________________||
def test_nested_loops(mock_create_presentation, wrap_end_pickup):

    disable()

    for i in atpbar(range(4)):
        for j in atpbar(range(3)):
            pass

    ## print()
    ## print(mock_create_presentation)

    assert 0 == wrap_end_pickup.call_count

##__________________________________________________________________||
def run_with_threading(nthreads=3, niterations=[5, 5, 5]):
    def task(n, name):
        for i in atpbar(range(n), name=name):
            time.sleep(0.0001)
    threads = [ ]
    for i in range(nthreads):
        name = 'thread {}'.format(i)
        n = niterations[i]
        t = threading.Thread(target=task, args=(n, name))
        t.start()
        threads.append(t)
    for t in threads:
        t.join()
    flush()

@pytest.mark.parametrize('niterations', [[5, 4, 3], [5, 0, 1], [0], [1]])
@pytest.mark.parametrize('nthreads', [3, 1, 0])
def test_threading(mock_create_presentation, wrap_end_pickup, nthreads, niterations):

    disable()

    # make niterations as long as nthreads. repeat if necessary
    niterations = list(itertools.chain(*itertools.repeat(niterations, nthreads//len(niterations)+1)))[:nthreads]

    run_with_threading(nthreads, niterations)

    ## print()
    ## print(mock_create_presentation)

    assert 1 == wrap_end_pickup.call_count

    presentations = mock_create_presentation.presentations
    assert 0 == len(presentations)


##__________________________________________________________________||
def run_with_multiprocessing(nprocesses, ntasks, niterations):
    def task(n, name):
        for i in atpbar(range(n), name=name):
            time.sleep(0.0001)
    def worker(reporter, task, queue):
        register_reporter(reporter)
        while True:
            args = queue.get()
            if args is None:
                queue.task_done()
                break
            task(*args)
            queue.task_done()
    reporter = find_reporter()
    queue = multiprocessing.JoinableQueue()
    for i in range(nprocesses):
        p = multiprocessing.Process(target=worker, args=(reporter, task, queue))
        p.start()
    for i in range(ntasks):
        name = 'task {}'.format(i)
        n = niterations[i]
        queue.put((n, name))
    for i in range(nprocesses):
        queue.put(None)
        queue.join()
    flush()

@pytest.mark.xfail()
@pytest.mark.parametrize('niterations', [[5, 4, 3], [5, 0, 1], [0], [1]])
@pytest.mark.parametrize('ntasks', [3, 1, 0])
@pytest.mark.parametrize('nprocesses', [4, 1])
def test_multiprocessing(mock_create_presentation, wrap_end_pickup, nprocesses, ntasks, niterations):

    disable()

    # make niterations as long as ntasks. repeat if necessary
    niterations = list(itertools.chain(*itertools.repeat(niterations, ntasks//len(niterations)+1)))[:ntasks]

    run_with_multiprocessing(nprocesses, ntasks, niterations)

    ## print()
    ## print(mock_create_presentation)

    assert 1 == wrap_end_pickup.call_count

    presentations = mock_create_presentation.presentations
    assert 0 == len(presentations)

##__________________________________________________________________||
